
local LngFile = mf.fsplit(select(1, ...),7) .. '*.lng'
local BAX_GetMsg = require"BAX.GetMsg"
local function GetMsg(tag)
 return BAX_GetMsg(LngFile, tag, true) --or '<Error>'
end
-------------------------------------------------------------------------------
--[[
  Макрос - заменитель плагина Active-Help
  © BAX, 19.11.2017 Версия 13
  Ctrl+F1 - вызов сопоставленной справки
  Ctrl+Enter - ShellExecute
  Alt+Shift+F12 - диалог настройки
--]]
local Info = package.loaded.regscript or function(...) return ... end
local nfo = Info {_filename or ...,
  name        = "Active-Help";
  description = GetMsg('description'); --"Заменитель Active-Help";
  version     = "1.13.5."; --http://semver.org/lang/ru/
  author      = "BAX";
  url         = "https://forum.farmanager.com/viewtopic.php?p=145176#p145176";
  id          = "619A960A-0DA9-4B6E-A7B1-684241C05EAF";
  history     = [[
08.10.2020 1.13.5
  1. настройка скрипта включена в меню настроек плагинов
  2. Исправлена одна мелкая ошибка
  3. Обновлены модули из комплекта поставки скрипта
  ]];
}
if not nfo then return end

local F = far.Flags
local tin = table.insert
local HistPref = 'BAX.Active-Help.'
local Version = 10

local TabUtils=require"Bax.TabUtils"
local SettingsKeys={Autor='BAX',
                    Settings='Active-Help',
                    Indexes='Active-Help.CHMIndexes'
                   }
local SI = {}
local wasErr=false



local C = far.Colors

local MenuColorIndex = {
  C.COL_MENUTEXT,             -- подложка
  C.COL_MENUBOX,              -- рамка
  C.COL_MENUTITLE,            -- заголовок - верхний и нижний
  C.COL_MENUTEXT,             -- Текст пункта
  C.COL_MENUHIGHLIGHT,        -- HotKey
  C.COL_MENUTEXT,             -- separator
  C.COL_MENUSELECTEDTEXT,     -- Выбранный
  C.COL_MENUSELECTEDHIGHLIGHT,-- Выбранный - HotKey
  C.COL_MENUSCROLLBAR,        -- ScrollBar
  C.COL_MENUDISABLEDTEXT,     -- Disabled
  C.COL_MENUARROWS,           -- '<' & '>' обычные
  C.COL_MENUARROWSSELECTED,   -- '<' & '>' выбранные
  C.COL_MENUARROWSDISABLED,   -- '<' & '>' Disabled
  C.COL_MENUGRAYTEXT,         -- "серый"
  C.COL_MENUSELECTEDGRAYTEXT, -- выбранный "серый"
}

local MenuPalette = {}
for i,v in ipairs(MenuColorIndex) do MenuPalette[i] = v and far.AdvControl(F.ACTL_GETCOLOR,v) end

----------------------------------------------
local function ExpandEnv(s)
 return s:gsub('%b%%', function(ss) return win.GetEnv(ss:sub(2,-2)) end) or s
end
----------------------------------------------

local function _FileExt(sName)
 return mf.fsplit(sName, 0x00000008):lower()
 --if sName:find('\\',1,true) then
 -- return (sName:match('.*%\\(.*)'):match('.*%.(.*)') or ''):lower()
 --else
 -- return (sName:match('.*%.(.*)') or ''):lower()
 --end
end

local function TableClone(Src, Res)
 local function _find(_t, _a)
  for _i,_ in pairs(_t) do
   if _t[_i][1] == _a then
    return _i
   end
  end
  return nil
 end
 if type(Res) ~= 'table' then
  Res = {}
 else
  -- Очистка
  for i,_ in pairs(Res) do
   Res[i] = nil
  end
 end
 local was={}
 for i,v in pairs(Src) do
  if type(v) == 'table' then
   local j=_find(was, v)
   if j then
    Res[i] = was[j][2]
   else
    local aa=TableClone(v)
    tin(was, {v,aa})
    Res[i] = aa
   end
  else
   Res[i] = v
  end
 end
 return Res
end

-- +
-- Настройки
local function DefSettings()
return{
  HelpKey="CtrlF1",
  ExecKey="CtrlEnter",
  SetupKey="ALtShiftF12",
  Prefix = 'achlp:ac-hlp',
  Prefixc = 'achlpc:ac-hlp-c',
  p7z_exe = '7z.exe',
  HelpTable={}
}
end

local S = mf.mload(SettingsKeys.Autor, SettingsKeys.Settings, 'roaming') or
          mf.mload(SettingsKeys.Autor, SettingsKeys.Settings, 'local') or
          {}

local function MergeSettings(S1, S2)
 for i,v in pairs(S2) do
  S1[i] = S1[i] or v
 end
 return S1
end

S = MergeSettings(S, DefSettings())

local HelpFiles, AutoSearch

local function ChmIndex(fn, bSave, p7z_exe)
 -- Функция составления, записи индекса и оглавления CHM-файла и
 -- поиска по нему.
 -- Функция преобразования ANSI в UTF-8
 local function ANSItoUTF_8(str)
   return win.WideCharToMultiByte(win.MultiByteToWideChar(str, win.GetACP()), 65001)
 end
 -- Функция сравнения для сортировки
 local function _SortProc(a1, a2)

  if a1.NameL < a2.NameL then return true
  elseif a1.NameL > a2.NameL then return false
  else -- a1.NameL == a2.NameL
   if a1.Name < a2.Name then return true
   elseif a1.Name > a2.Name then return false
   else -- a1.Name == a2.Name
    if a1.N < a2.N then return true end
   end
  end
  return false
 end -- _SortProc
 -- Преобразование строки в "строчный с заглавной буквы"
 -- (чтоб при занесении в индекс не зависеть от исходного файла)
 local function capital(s)
  return mf.ucase(mf.substr(s,0,1))..mf.lcase(mf.substr(s,1))
 end
 --------------- начало выполнения ChmIndex ----------------------
 for i=1,#SI do
  if SI[i].File == fn then
   if SI[i].FDate == win.GetFileInfo(fn).LastWriteTime then
    return SI[i].Index
   else
    table.remove(SI, i)
    break
   end
  end
 end
 -- Сюда попадаем, если либо индекса ещё не было, либо он неактуален

 p7z_exe = p7z_exe or S.p7z_exe or '7z.exe'
 if (not p7z_exe) or (not mf.fexist(p7z_exe)) then
  p7z_exe = win.SearchPath(win.GetEnv('path'), '7z', '.exe')
  if not p7z_exe then
   if not wasErr then
    far.Show('Не найдена программа 7z.exe\nПоиск по файлу chm невозможен','Active-Help', nil, 'w')
    wasErr = true
    return -- ChmIndex
   end
  end
 end

 local aIndex, Ind1, aInd2={}, nil,{}
 local ff

 for _,v in ipairs{'hhk','hhc'} do
  ff = io.popen('""'..p7z_exe..'" e -so "'..fn..'" *.'..v..'"')
  if ff then
   ff:seek('set',0)
   local iLine=0
   for Line in ff:lines() do
    iLine = iLine + 1
    Line = mf.trim(ANSItoUTF_8(Line))
    if mf.index(Line, '<LI>') == 0 then Line = mf.substr(Line, 4) end
    if Line:lower():match('<object type="text/sitemap">') then
     Ind1 = {}
    elseif mf.index(Line:lower(), '</object>') == 0 and Ind1 and
           Ind1.Local and Ind1.Name   then
     local Keyword = Ind1.Keyword
     Ind1.Keyword = nil
     Ind1.N=#aIndex+1
     table.insert(aIndex, Ind1)
     if Keyword and Keyword ~= Ind1.Name then
      iLine = iLine + 1
      local Ind2 = TableClone(Ind1)
      Ind2.N = iLine
      Ind2.Name = Keyword
      Ind2.NameL = Keyword:lower()
      table.insert(aIndex, Ind2)



     end
     Ind1 = nil
    elseif Ind1 ~= nil and mf.index(Line, '<param name=') == 0 then
     local sPar, sVal=Line:match('name="(%w*)" value="(.*)"')
     if sPar and mf.index('local name keyword', sPar) >= 0 and mf.trim(sVal) ~= '' then
      sPar = capital(sPar)
      sVal = mf.trim(sVal)
      Ind1[sPar] = sVal
      if sPar =='Name' then
       --Ind1.NameL = mf.lcase(sVal)
       ---- Теперь заодно вычленяю первое слово
       ----Ind1.NameL = sVal:lower():match('[^%w_]*([%w_]+).*') or sVal:lower()
       -- Теперь заодно удаляю всё, кроме букв и символов .:-/
       Ind1.NameL = mf.trim(sVal:lower():match('[^%w_]*([%w_%.:%-%/%s]+).*') or sVal:lower())
      elseif sPar == 'Keyword' then
       Ind1.Keyword = sVal
      end
     end
    end
   end;
   ff:close()
  else -- not ff
   far.Message("Неудача распаковки индеса\nВозможно, не установлен 7-zip", "Active-Help", nil, 'w' )
   exit()
  end -- if ff
  if #aIndex > 0 then
   table.sort(aIndex, _SortProc)
   aInd2[v] = aIndex
  end
 end -- for _,v in ipairs{'hhk','hhc'} do

 tin(SI,{File=fn,
         FDate=win.GetFileInfo(fn).LastWriteTime,
         Index=aInd2
        })
 if bSave then
  mf.msave(SettingsKeys.Autor, SettingsKeys.Indexes, SI)
 end
 return aInd2
end -- ChmIndex
-- Процедура построения ВСЕХ индексов
local function CreateAllIndexex()
 local fn
 for _,v in pairs(S.HelpTable) do
  for _,vv in pairs(v.Helps) do
   fn = ExpandEnv(vv.File)
   if _FileExt(fn) == '.chm' then
    far.Message('Индексация '..vv.File..'\n('..(vv.Alias or 'без алиаса')..')','','')
    if not ChmIndex(fn, false) then break end
    far.AdvControl("ACTL_REDRAWALL")
   end
  end
 end
 far.Message('Сохранение индексов','','')
 mf.msave(SettingsKeys.Autor, SettingsKeys.Indexes, SI)
 far.AdvControl("ACTL_REDRAWALL")

end -- CreateAllIndexex


-- Фунция бинарного поиска. Case Sensitive, увы! Как сделать регистронезвисимый поиск - пока не придумал
local function BinSearch(aIndex, AsKey, lCaseSensitive)

  local i1, i2, i=1,#aIndex,-1
  local sKey = mf.lcase(AsKey)
  -- Поиск регистронезависимый
  while i2 > i1 do
   if i2-i1==1 then
    if aIndex[i1].NameL == sKey then
     i=i1
    elseif aIndex[i2].NameL == sKey then
     i=i2
    else
     i = -1
    end
    break
   end
   i = math.floor( (i1+i2)/2 )
   if aIndex[i].NameL < sKey then
    i1 = i
   elseif aIndex[i].NameL > sKey then
    i2 = i
   else
    while i>1 and aIndex[i-1].NameL == sKey do
     i = i - 1
    end
    break
   end
  end
  if i > 0 then
   if not lCaseSensitive then
    return aIndex[i].Local
   else
    -- Регистрозависимый поиск. Уточнение перебором
    for ii=i, #aIndex do
     if aIndex[ii].Name > AsKey then
      return
     elseif aIndex[ii].Name == AsKey then
      return aIndex[ii].Local
     end
    end
   end
  end
end -- BinSearch

local function OpnChm(fn, SearchStr, iSearch)
 local res=nil
 local idx=ChmIndex(fn, true)
 if idx then
  -- 1. Поиск точного совпадения
  if (iSearch==nil) or (iSearch == 1) then
   for _,v in ipairs({'hhk','hhc'}) do
    if idx[v] then
     res = BinSearch(idx[v], SearchStr, true)
     if res then
      return res
     end
    end
   end
  end
  -- 2. Регистронезависимый поиск
  if (iSearch==nil) or (iSearch == 2) then
   for _,v in ipairs({'hhk','hhc'}) do
    if idx[v] then
     res = BinSearch(idx[v], SearchStr, false)
     if res then
      return res
     end
    end
   end
  end
  -- 3. "частичный" поиск (любого слова). Например, при полиске mf.index найдёт mf
  if (iSearch==nil) or (iSearch == 3) then
   for _,v in ipairs({'hhk','hhc'}) do
    if idx[v] then
     for w in SearchStr:gmatch('([%w-]+)') do
      res =  BinSearch(idx[v], w, false)
      if res then
       return res
      end
     end
    end
   end
  end
 end
 return res
end -- OpnChm

local function _InKey(_s)
 far.Message(_s.."\nPress key","",'')
 local res=mf.waitkey()
 far.AdvControl("ACTL_REDRAWALL")
 return res
end -- _InKey

local function _OneExtSetup(AHelpTable)
 local HelpTable = TableClone(AHelpTable)
 local DlgGUID = win.Uuid('7F41CA3C-03CE-49FB-82AE-2D33C0A1927D')
 local DlgItems={}
 local ID={}
 local y
 local busy=false
 tin(DlgItems, { "DI_DOUBLEBOX", 0, 0, 73, 25, 0, 0, 0, 0, 'Редактирование помощи' } )
 y = 2
 tin(DlgItems, { "DI_TEXT", 3, y, 3, y, 0,0,0,0, '&1. Маска файла'} )
 tin(DlgItems, { "DI_EDIT", 20, y, 68, y, 0, HistPref.."Mask",0,{DIF_HISTORY=1, DIF_SELECTONENTRY=1}, HelpTable.Mask})
 ID.Mask = #DlgItems
 y = y + 2
 local lbList = {}
 for i=1,#HelpTable.Helps do
  lbList[i] = {Text=HelpTable.Helps[i].File}
 end
 tin( DlgItems, {"DI_TEXT", 0, y, 0, y, 0,0,0,"DIF_CENTERGROUP", "&2. Список файлов  "} )
 y = y + 1
 tin( DlgItems, {"DI_LISTBOX", 3, y, 68, y+5, lbList, 0, 0,
                 "DIF_LISTNOCLOSE", ""} )
 ID.ListBox = #DlgItems
 y = y + 6
 tin( DlgItems, {"DI_TEXT", 3,y,32,y,0,0,0,0,"&3. Алиас файла помощи"} )
 tin( DlgItems, {"DI_EDIT", 35,y,68,y,0,0,0,"DIF_SELECTONENTRY",#HelpTable.Helps > 0 and HelpTable.Helps[1].Alias or ""} )
 ID.Alias = #DlgItems
 y = y + 1

 tin( DlgItems, {"DI_TEXT", 3,y,32,y,0,0,0,0,"&3. Команда для вызова помощи"} )
 tin( DlgItems, {"DI_EDIT", 35,y,68,y,0,HistPref..'Cmd',0,{DIF_HISTORY=1, DIF_SELECTONENTRY=1},#HelpTable.Helps > 0 and HelpTable.Helps[1].Cmd or ""} )
 ID.Cmd = #DlgItems
 y = y + 1

 tin( DlgItems, {"DI_TEXT", 3,y,37,y+1,0,0,0,"DIF_WORDWRAP","&4. Параметры команды\n   (%s=отмеченный текст, файл)"} )
 tin( DlgItems, {"DI_EDIT", 35,y,68,y,0,HistPref.."Par",0,{DIF_HISTORY=1, DIF_SELECTONENTRY=1}, #HelpTable.Helps > 0 and HelpTable.Helps[1].Par or ""} )
 ID.Par = #DlgItems

 y = y + 2
 tin( DlgItems, {"DI_TEXT", -1 ,y,73,y, 0,0,0,"DIF_SEPARATOR", ''} )
 y = y + 1
 tin( DlgItems, {"DI_BUTTON", 0,y,0,y,0,0,0,{DIF_DEFAULTBUTTON=1,DIF_CENTERGROUP=1}, 'Ok'} )
 ID.Ok = #DlgItems
 tin( DlgItems, {"DI_BUTTON", 0,y,0,y,0,0,0,"DIF_CENTERGROUP", 'Cancel'} )
 ID.Cancel = #DlgItems
 y = y + 2
 ---------------------------------------------------------------------------
 local function _DlgProc(hDlg, Msg, ItemId, Param2)
  -- Функция сохранения данных
  local function _SaveOneHelp()
   local ii = hDlg:send("DM_LISTGETCURPOS", ID.ListBox).SelectPos
   if ii == 0 then return end
   local s
   s = mf.trim( hDlg:send("DM_GETTEXT", ID.Alias) )
   if s ~= '' then
    HelpTable.Helps[ii].Alias = s
   else
    HelpTable.Helps[ii].Alias = nil
   end
   s = mf.trim( hDlg:send("DM_GETTEXT", ID.Cmd) )
   if s ~= '' then
    HelpTable.Helps[ii].Cmd = s
    s = hDlg:send("DM_GETTEXT", ID.Par)
    if s ~= '' then
     HelpTable.Helps[ii].Par = s
    else
     HelpTable.Helps[ii].Par = nil
    end
   else
    HelpTable.Helps[ii].Cmd = nil
    HelpTable.Helps[ii].Par = nil
   end
  end -- _SaveOneHelp
  if Msg == F.DN_CTLCOLORDLGLIST then
   return MenuPalette
  elseif Msg == F.DN_INITDIALOG then
   hDlg:send("DM_LISTSETTITLES", ID.ListBox, {Bottom="Ins - новый; Del - удалить; F4 - поменять"})
  elseif Msg == F.DN_LISTCHANGE then
   if not busy then
    busy = true
    --   CurrPos - Старая позиция; Param2 -- Новая позиция
    -- Сначала надо сохранить то, что в старую позицию занести;
    _SaveOneHelp()
    -- Затем вывести из новой позиции
    if Param2 > #HelpTable.Helps then
     hDlg:send("DM_SETTEXT", ID.Alias, "")
     hDlg:send("DM_SETTEXT", ID.Cmd, "")
     hDlg:send("DM_SETTEXT", ID.Par, "")
    else
     if HelpTable.Helps[Param2].Cmd then
      hDlg:send("DM_SETTEXT", ID.Cmd, HelpTable.Helps[Param2].Cmd)
      if HelpTable.Helps[Param2].Par then
       hDlg:send("DM_SETTEXT", ID.Par, HelpTable.Helps[Param2].Par)
      else
       hDlg:send("DM_SETTEXT", ID.Par, "")
      end
     else
      hDlg:send("DM_SETTEXT", ID.Cmd, "")
      hDlg:send("DM_SETTEXT", ID.Par, "")
     end
     if HelpTable.Helps[Param2].Alias then
      hDlg:send("DM_SETTEXT", ID.Alias, HelpTable.Helps[Param2].Alias)
     else
      hDlg:send("DM_SETTEXT", ID.Alias, "")
     end
    end
    busy = false
   end
  elseif Msg == F.DN_CONTROLINPUT and
         Param2.EventType == F.KEY_EVENT and
         ItemId == ID.ListBox then
   local _key=far.InputRecordToName(Param2)
   local aListPos = hDlg:send("DM_LISTGETCURPOS", ID.ListBox)
   local iLine=aListPos.SelectPos
   if _key == 'Up' and hDlg:send("DM_LISTGETCURPOS", ID.ListBox).SelectPos == 1 then
    -- Поиск предыдущего фокусируемого контрола
    hDlg:send("DM_SETFOCUS", ID.Mask)
    return true
   elseif _key == 'Down' and
          iLine== hDlg:send("DM_LISTINFO",ID.ListBox).ItemsNumber then
    hDlg:send("DM_SETFOCUS", ID.Alias)
    return true
   elseif _key == "Del" and iLine > 0 and iLine <= #HelpTable.Helps then
    if HelpTable.Helps[iLine].Deleted then
     HelpTable.Helps[iLine].Deleted = nil
    else
     HelpTable.Helps[iLine].Deleted = true
    end
    local CurrItem=hDlg:send("DM_LISTGETITEM", ID.ListBox, iLine)
    if HelpTable.Helps[iLine].Deleted then
     busy = true
     hDlg:send("DM_LISTUPDATE",
               ID.ListBox,
               {Index=iLine,
                Text=CurrItem.Text .. ' (удалён)',
                Flags=bor(CurrItem.Flags,F.LIF_GRAYED)
               } )
     hDlg:send("DM_LISTSETCURPOS", ID.ListBox, aListPos)
     busy = false
    else
     busy = true
     hDlg:send("DM_LISTUPDATE",
               ID.ListBox,
               {Index=iLine,
                Text=mf.substr(CurrItem.Text, 0, mf.len(CurrItem.Text)-9)
                , Flags=bxor(CurrItem.Flags,F.LIF_GRAYED)
               } )
     hDlg:send("DM_LISTSETCURPOS", ID.ListBox, aListPos)
     busy = false
    end
   elseif _key == "Ins" then
    local s = far.InputBox(nil, 'Файл помощи', 'Введите имя файла помощи', nil,
                           _key == 'F4' and hDlg:send("DM_GETTEXT", ID.ListBox) or '',
                           nil, nil, {FIB_EDITPATH=1, FIB_BUTTONS=1})
    if s and s ~= '' then
     tin(HelpTable.Helps, {File=s} )
     hDlg:send("DM_LISTADDSTR", ItemId, s)
     hDlg:send("DM_LISTSETCURPOS", ID.ListBox, {TopPos = -1, SelectPos = #HelpTable.Helps})
    end
   elseif _key=='F4' then
    local s = far.InputBox(nil, 'Файл помощи', 'Введите имя файла помощи', nil,
                           _key == 'F4' and hDlg:send("DM_GETTEXT", ID.ListBox) or '',
                           nil, nil, {FIB_EXPANDENV=1, FIB_EDITPATH=1, FIB_BUTTONS=1})
    if s and s ~= '' then
     HelpTable.Helps[hDlg:send("DM_LISTGETCURPOS", ID.ListBox).SelectPos].File = s
     hDlg:send("DM_SETTEXT", ItemId, s)
    end
   elseif _key:match('^Alt%w$') then
    -- Увы, не выйти по горячей клавише из ListBox-а ... Сам поищу
    local i, aItem=0
    repeat
     i = i + 1
     aItem = hDlg:send("DM_GETDLGITEM", i)
     if aItem then
      if mf.index(aItem[10], '&'..mf.substr(_key, 3,1)) >= 0 then
       if hDlg:send("DM_SETFOCUS", i) == 0 then
        hDlg:send("DM_SETFOCUS", i+1)
       end
       break
      end;
     end
    until not aItem
    return true
   end
  elseif Msg == F.DN_CLOSE and ItemId == ID.Ok then
   HelpTable.Mask = hDlg:send("DM_GETTEXT", ID.Mask)
   _SaveOneHelp()
  end
 end -- _DlgProc

 local y0 = mf.max(1, mf.min(15, Far.Height-y-1))
 if far.Dialog(DlgGUID,-1,y0,73,y0-1+y, nil, DlgItems, "FDLG_SMALLDIALOG", _DlgProc) == ID.Ok then
  return HelpTable
 else
  return nil
 end
end -- _OneExtSetup

local function Active_Help_Setup()
  local DlgItems, ID, y = {}, {}
  local DlgGUID = win.Uuid('B8C5FF7A-9195-4096-AD60-0541E2D23102')
  local SS = TableClone(S)
  local w = math.max(73,math.floor(Far.Width*0.75+0.5))
  tin(DlgItems, { "DI_DOUBLEBOX", 0, 0, w, 15, 0, 0, 0, 0, GetMsg('SetupTitle')}) --'Настройка макроса - заменителя Active-Help' })
  y = 2
  tin(DlgItems, { "DI_TEXT", 3, y, 0, y, 0, 0, 0, 0, '&1. '..GetMsg('ExecKey')}) --'Клавиша для выполнения команды' })
  tin(DlgItems, { "DI_EDIT", 37, y, 55, y, 0, "", 0, "DIF_SELECTONENTRY", SS.ExecKey })
  ID.ExecKey = #DlgItems
  tin(DlgItems, {"DI_BUTTON", 57, y, 57, y, 0,0,0, {DIF_BTNNOCLOSE=1, DIF_NOFOCUS=1}, "Key"})
  ID.ExecKeyClick = #DlgItems

  y = y + 1
  tin(DlgItems, { "DI_TEXT", 3, y, 0, y, 0, 0, 0, 0, '&2. '..GetMsg('HelpKey') }) -- 'Клавиша для вызова помощи'
  tin(DlgItems, { "DI_EDIT", 37, y, 55, y, 0, "", 0, "DIF_SELECTONENTRY", SS.HelpKey })
  ID.HelpKey=#DlgItems
  tin(DlgItems, {"DI_BUTTON", 57, y, 57, y, 0,0,0, {DIF_BTNNOCLOSE=1, DIF_NOFOCUS=1}, "Key"})
  ID.HelpKeyClick = #DlgItems

  y = y + 1
  tin(DlgItems, {"DI_TEXT", 2,y,w-3,y+1,0,0,0,{DIF_WORDWRAP=1, DIF_CENTERTEXT=1}, '&3. '..GetMsg('CmdlPrefixes')}) -- 'Префиксы командной строки (без пробелов; допускается несколько перфиксов через ":")'
  y = y + (DlgItems[#DlgItems][10]:len() > w-5 and 2 or 1)
  do
   local l,n,k=0,0,#DlgItems+1
   tin(DlgItems, {"DI_TEXT", 7,y,0,y,0,0,0,0, GetMsg('AutoSelect')}); l=l+DlgItems[#DlgItems][10]:len() -- 'автовыбор:'
   tin(DlgItems, {"DI_EDIT", 19,y,35,y,0,0,0,{DIF_SELECTONENTRY=1}, SS.Prefix}); l=l+DlgItems[#DlgItems][4]-DlgItems[#DlgItems][2]+1; n=n+1
   ID.Prefix = #DlgItems
   tin(DlgItems, {"DI_TEXT", 37,y,0,y,0,0,0,0, GetMsg('ManualSelect')});l=l+DlgItems[#DlgItems][10]:len(); n=n+1 -- 'с выбором файла:'
   tin(DlgItems, {"DI_EDIT", DlgItems[#DlgItems][2] +DlgItems[#DlgItems][10]:len()+1,y,70,y,0,0,0,{DIF_SELECTONENTRY=1}, SS.Prefixc}); l=l+DlgItems[#DlgItems][4]-DlgItems[#DlgItems][2]+1; n=n+1
   --tin(DlgItems, {"DI_EDIT", 50,y,70,y,0,0,0,{DIF_SELECTONENTRY=1}, SS.Prefixc})
   ID.Prefixc = #DlgItems
   l = l + n*2 -- 2 пробела между текстами. l тут - общая ширина текстов и полей ввода
   n = math.ceil((w - l) / 2)
   --for i=k,#DlgItems do
   -- if DlgItems[i][1] == 'DI_EDIT' or DlgItems[i][1] == F.DI_EDIT then
   --end
  end
  y = y + 1
  tin(DlgItems, { "DI_TEXT", 3, y, 0, y, 0, 0, 0, 0, '&4. Клавиша для вызова настройки' })
  tin(DlgItems, { "DI_EDIT", 37, y, 55, y, 0, "", 0, "DIF_SELECTONENTRY", SS.SetupKey or '' })
  ID.SetupKey=#DlgItems
  tin(DlgItems, {"DI_BUTTON", 57, y, 57, y, 0,0,0, {DIF_BTNNOCLOSE=1, DIF_NOFOCUS=1}, "Key"})
  ID.SetupKeyClick = #DlgItems

  local lbList={}
  for i=1,#SS.HelpTable do
   tin(lbList, {Text=SS.HelpTable[i].Mask})
  end
  tin(DlgItems, {"DI_DOUBLEBOX", 1,y+1,w-2,y+7,0,0,0,0,"&5. Настройка файлов помощи"})
  ID.MaskListBox = #DlgItems
  tin(DlgItems, {"DI_LISTBOX", 2,y+2,w-3,y+6,
                  lbList, 0, 0,
                  {DIF_LISTNOCLOSE=1,DIF_LISTNOBOX=1}, ""})
  ID.MaskList = #DlgItems
  y = y + 7
  tin(DlgItems, { "DI_BUTTON", 0,y,0,y,0,0,0,{DIF_BTNNOCLOSE=1,DIF_CENTERGROUP=1, DIF_NOFOCUS=1},'Редакт. (F4)'})
  ID.BtnEdit = #DlgItems
  tin(DlgItems, { "DI_BUTTON", 0,y,0,y,0,0,0,{DIF_BTNNOCLOSE=1,DIF_CENTERGROUP=1, DIF_NOFOCUS=1},'Удалить (F8)'})
  ID.BtnDel = #DlgItems
  tin(DlgItems, { "DI_BUTTON", 0,y,0,y,0,0,0,{DIF_BTNNOCLOSE=1,DIF_CENTERGROUP=1, DIF_NOFOCUS=1},'Добавить (F2)'} )
  ID.BtnAdd = #DlgItems

  y = y + 1
  tin(DlgItems, {"DI_TEXT", 3,y,0,y,0,0,0,0,'&6. Путь к программе 7z.exe'})
  tin(DlgItems, {"DI_EDIT", 35,y,w-3,y,0,0,0,{DIF_EDITPATH=1, DIF_EDITPATHEXEC=1, DIF_SELECTONENTRY=1},
                 (SS.p7z_exe or '')=='' and '7z.exe' or SS.p7z_exe})
  ID.p7z = #DlgItems

  y = y + 1
  tin(DlgItems, {"DI_TEXT", 3, y, w-3, y, 0,0,0, 0, "&7. Автоматический поиск по всем chm-файлам"})
  tin(DlgItems, {"DI_CHECKBOX", 46, y, w-3, y, S.AutoSearch and 1 or 0,1, 1, 0, ""})
  --tin(DlgItems, {"DI_CHECKBOX", 46, y, 70, y, 0,0,S.AutoSearch and 1 or 0, 0, ""})
  ID.AutoSearch = #DlgItems

  y = y + 1
  tin(DlgItems, { "DI_TEXT", -1 ,y,w,y, 0,0,0, "DIF_SEPARATOR", ''} )
  y = y + 1
  tin(DlgItems, { "DI_BUTTON", 0, y, 0, y, 0, 0,0, {DIF_DEFAULTBUTTON=1,DIF_CENTERGROUP=1},"Сохранить" })
  ID.SaveBtn = #DlgItems
  tin(DlgItems, { "DI_BUTTON", 0, y, 0, y, 0, 0,0, "DIF_CENTERGROUP","Отмена" })
  tin(DlgItems, { "DI_BUTTON", 0, y, 0, y, 0, 0,0, {DIF_CENTERGROUP=1},"Сброс" })
  ID.Reset = #DlgItems
  -- Проиндексировать и сохранить индексы
  y = y + 1
  tin(DlgItems, { "DI_BUTTON", 0, y, 0, y, 0, 0,0, {DIF_CENTERGROUP=1, DIF_BTNNOCLOSE=1},"Импорт" })
  ID.ImportBtn = #DlgItems
  tin(DlgItems, { "DI_BUTTON", 0, y, 0, y, 0, 0,0, {DIF_CENTERGROUP=1, DIF_BTNNOCLOSE=1},"Экспорт" })
  ID.ExportBtn = #DlgItems
  -- Проверка существования 7z.exe
  local p7z_exe = S.p7z_exe or ''
  if (p7z_exe == '') or (not mf.fexist(p7z_exe)) then p7z_exe = '7z.exe' end
  if p7z_exe and p7z_exe ~= '' and win.SearchPath(win.GetEnv('path'), '7z.exe') then
   tin(DlgItems, { "DI_BUTTON", 0, y, 0, y, 0, 0,0, {DIF_CENTERGROUP=1, DIF_BTNNOCLOSE=1},"Индексировать" })
  else
   tin(DlgItems, { "DI_BUTTON", 0, y, 0, y, 0, 0,0, {DIF_CENTERGROUP=1, DIF_BTNNOCLOSE=1, DIF_DISABLE=1},"Индексировать" })
  end
  ID.CreateIndexes = #DlgItems
  DlgItems[1][5] = y+3
  --------------------------------------------------------
  local function _DlgProc(hDlg, Msg, ItemId, Param2)
   local function _SaveHelps()
    SS.ExecKey = hDlg:send("DM_GETTEXT", ID.ExecKey)
    SS.HelpKey = hDlg:send("DM_GETTEXT", ID.HelpKey)
    SS.SetupKey = hDlg:send("DM_GETTEXT", ID.SetupKey)
    SS.Prefix = hDlg:send("DM_GETTEXT", ID.Prefix)
    SS.Prefixc = hDlg:send("DM_GETTEXT", ID.Prefixc)
    p7z_exe = hDlg:send("DM_GETTEXT", ID.p7z)
    SS.p7z_exe = p7z_exe
   end
   local function _ProcessOneHelp(iLine, bEdit)
    if iLine >= 0 and iLine <= #SS.HelpTable then
     if bEdit then
      local New
      hDlg:send("DM_SHOWDIALOG", 0)
      if iLine > 0 then
       New = _OneExtSetup(SS.HelpTable[iLine])
      else
       New = _OneExtSetup({Mask='', Helps={}})
      end
      hDlg:send("DM_SHOWDIALOG", 1)
      if New then
       for _i=#New.Helps, 1, -1 do
        if New.Helps[_i].Deleted then
         table.remove(New.Helps, _i)
        end
       end
       if iLine == 0 then
        tin(SS.HelpTable, New)
        hDlg:send( "DM_LISTADDSTR", ID.MaskList, New.Mask)
        hDlg:send( "DM_LISTSETCURPOS", ID.MaskList, {TopPos=-1, SelectPos=#SS.HelpTable} )
       else
        SS.HelpTable[iLine] = New
        hDlg:send( "DM_SETTEXT", ID.MaskList, New.Mask)
       end
      end
     else -- not bEdit
      local CurrItem=hDlg:send("DM_LISTGETITEM", ID.MaskList, iLine)
      SS.HelpTable[iLine].Deleted = not SS.HelpTable[iLine].Deleted
      hDlg:send("DM_LISTUPDATE", ID.MaskList,
                {Index=iLine,
                 Text=CurrItem.Text,
                 Flags=bxor(CurrItem.Flags, F.LIF_GRAYED)
              } )
     end
    end
   end -- _ProcessOneHelp
   local Res=nil --0
   local ItemData=far.GetDlgItem(hDlg, ItemId)
   if Msg == F.DN_GOTFOCUS then
    if ItemId == ID.MaskListBox then
     hDlg:send( "DM_SETFOCUS", ID.MaskList)
    end
   elseif Msg == F.DN_EDITCHANGE then
    if ItemId == ID.p7z then
     hDlg:send("DM_ENABLE", ID.CreateIndexes,
               win.SearchPath(win.GetEnv('path'), hDlg:send("DM_GETTEXT", ID.p7z)) and 1 or 0 )
    end
   elseif Msg == F.DN_CTLCOLORDLGLIST then
    return MenuPalette
   elseif Msg == F.DN_BTNCLICK then
    if band(ItemData[9], F.DIF_BTNNOCLOSE) == F.DIF_BTNNOCLOSE then
     Res = 0
    end
    local function _ValidKey(k)
     local res={regex.match(k, '(Ctrl|Shift|Alt)?(Esc|Left|Right|Up|Down|Tab|PgDn|PgUp|Home|End)')}
     return not ( res[#res] or
                 (k == 'Enter') or
                 (k:len() == 1)
                )
    end -- _ValidKey
    if ItemId == ID.AutoSearch then
     SS.AutoSearch = hDlg:send("DM_GETCHECK", ID.AutoSearch) == F.BSTATE_CHECKED
    elseif ItemId == ID.ExecKeyClick then
     local k=_InKey('Задайте клавишу для выполнения команды')
     if not _ValidKey(k) then
      if k ~= 'Esc' then
       far.Message('Недопустимая клавиша '..k, 'Ошибка', ';Ok', 'w')
      end
     else
      hDlg:send("DM_SETTEXT", ID.ExecKey, k)
     end
    elseif ItemId == ID.HelpKeyClick then
     local k=_InKey('Задайте клавишу для вызова справки')
     if not _ValidKey(k) then
      if k ~= 'Esc' then
       far.Message('Недопустимая клавиша '..k, 'Ошибка', ';Ok', 'w')
      end
     else
      hDlg:send("DM_SETTEXT", ID.HelpKey, k)
     end
    elseif ItemId == ID.SetupKeyClick then
     local k=_InKey('Задайте клавишу для вызова диалога настройки')
     if not _ValidKey(k) then
      if k ~= 'Esc' then
       far.Message('Недопустимая клавиша '..k, 'Ошибка', ';Ok', 'w')
      end
     else
      hDlg:send("DM_SETTEXT", ID.SetupKey, k)
     end
    elseif ItemId == ID.BtnEdit or
           ItemId == ID.BtnDel or
           ItemId == ID.BtnAdd
    then
     local iLine
     if ItemId == ID.BtnAdd then
      iLine = 0
     else
      iLine=hDlg:send("DM_LISTGETCURPOS", ID.MaskList).SelectPos
     end
     _ProcessOneHelp(iLine, ItemId == ID.BtnEdit or ItemId == ID.BtnAdd)
    elseif ItemId == ID.SaveBtn then
     _SaveHelps()
    elseif ItemId == ID.Reset then
     if far.Message('После сброса восстановление данных будет невозможно!\nПродолжить?',
                    'Сброс настроек Active-Help',
                    ';OkCancel', 'w') == 1 then
      S = DefSettings()
      mf.mdelete(SettingsKeys.Autor, SettingsKeys.Settings)
      SI = {}
      mf.mdelete(SettingsKeys.Autor, SettingsKeys.Indexes)
     end
    elseif ItemId == ID.ImportBtn then
     local fn
     local Done=false
     repeat
      fn =far.InputBox(nil, 'Импорт настроек', 'Введите имя файла, откуда загрузить настройки', nil,
                       '',
                       nil, nil, {FIB_EXPANDENV=1, FIB_EDITPATH=1, FIB_BUTTONS=1})
      if not fn then break end
      if mf.fexist(fn) then
       local ImportProc=loadfile(fn)
       local ss=ImportProc and ImportProc()
       if type(ss) ~= 'table' then
        far.Message('Ошибка загрузки настроек', "Импорт настроек")
       else
        S = ImportProc()
        mf.msave(SettingsKeys.Autor, SettingsKeys.Settings, S)
        hDlg:send("DM_CLOSE", ID.Cancel)
        return
       end
      else
        far.Message('Файл не найден!', "Импорт настроек")
      end;
     until Done
    elseif ItemId == ID.ExportBtn then
     local fn=far.InputBox(nil, 'Экспорт настроек', 'Введите имя файла, куда выгрузить настройки', nil,
                           '',
                           nil, nil, {FIB_EXPANDENV=1, FIB_EDITPATH=1, FIB_BUTTONS=1})
     if not mf.fexist(fn) or far.Message('Файл\n'..fn..'\nсуществует. Переписать?', 'Экспорт настроек', ';YesNo')==1 then
      fn = io.open(fn, 'w')
      if fn then
       fn:write('return\n', TabUtils.TableToString(SS, ';\n'))
       fn:close()
      end
     end
    elseif ItemId == ID.CreateIndexes then
     if far.Message('Построение и сохранение индексов может значительно ускорить поиск по ключу.\n'..
                    'Однако процедура индексации может занять значительное время.\n\n'..
                    'Продолжить?',
                    'Построение индексов для CHM-файлов',
                    ';YesNo') == 1 then
      if far.Message('Принудительно обновить ВСЕ индексы?\n',
                     'Построение индексов для CHM-файлов',';YesNo','w')==1 then
       SI = {}
      end
      far.Message('Построение индексов...',nil,'')
      CreateAllIndexex()
      far.AdvControl("ACTL_REDRAWALL")
     end
    end -- if ItemId
   -- EndIF Msg == F.DN_BTNCLICK
   elseif Msg == F.DN_CONTROLINPUT then
    if Param2.EventType == F.KEY_EVENT then
     local _key=far.InputRecordToName(Param2)
     if _key == 'F4' or _key == 'F8' or _key == 'F2' or
        (_key == 'Del' and ItemId == ID.MaskList) or
        (_key == 'Ins' and ItemId == ID.MaskList) then
      local iLine
      if _key == 'Ins' or _key == 'F2' then
       iLine = 0
      else
       iLine=hDlg:send("DM_LISTGETCURPOS", ID.MaskList).SelectPos
      end
      _ProcessOneHelp(iLine, _key == 'F4' or _key == 'Ins' or _key == 'F2')
     elseif (_key=='Up' or _key == 'Down') and ItemId == ID.MaskList then
      local nn, jj = hDlg:send("DM_LISTINFO",ID.MaskList).ItemsNumber,
                     hDlg:send("DM_LISTGETCURPOS", ID.MaskList).SelectPos
      if _key == 'Up' and jj <= 1 then
       hDlg:send("DM_SETFOCUS", ID.SetupKey)
       return true
      elseif _key == 'Down' and jj == nn then
       hDlg:send("DM_SETFOCUS", ID.SaveBtn)
       return true
      end
     end
    end
   elseif Msg == F.DN_CLOSE and ItemId == ID.SaveBtn then
    _SaveHelps()
   end -- case по Msg
   return Res
  end -- _DlgProc
  if far.Dialog(DlgGUID, -1, 3, w, y+4, nil, DlgItems, "FDLG_SMALLDIALOG", _DlgProc) == ID.SaveBtn then
   -- Сохранение данных в память
   for i=#SS.HelpTable, 1, -1 do
    if SS.HelpTable[i].Deleted then
     table.remove(SS.HelpTable, i)
    end
   end
   S = SS
   S.Version = Version -- На будущее...
   mf.msave(SettingsKeys.Autor, SettingsKeys.Settings, S)
   far.MacroLoadAll()
  end
end -- Active_Help_Setup()

local function HelpFList(fn)
  -- Составление списка настроек на помощь, подходящих к типу файла
  HelpFiles = {}
  AutoSearch = S.AutoSearch -- Если в настройке автопоиск разрешён, то он включается только если все файлы - chm
  if #S.HelpTable > 0 then
   for i=1,#S.HelpTable do
    if fn == '*.*' or far.ProcessName(F.PN_CMPNAMELIST, S.HelpTable[i].Mask, fn, F.PN_SKIPPATH) then
     for j=1,#S.HelpTable[i].Helps do
      tin(HelpFiles, S.HelpTable[i].Helps[j])
      AutoSearch = AutoSearch and (_FileExt(S.HelpTable[i].Helps[j].File)=='.chm')
     end
    end
   end
  end
  return (#HelpFiles > 0)
end -- HelpFList

local function GetHelp(KeyString, AAutoSearch)
  local i,cmd,par,hf=0
  if not KeyString then return end
  if AAutoSearch and AutoSearch and #HelpFiles > 1 then
   -- Автоматический поиск по всем chm-файлам
   cmd = 'hh'
   for iSearch=1,3 do
    for _,hf1 in ipairs(HelpFiles) do
     par = OpnChm(ExpandEnv(hf1.File), KeyString, iSearch)
     if par then
      hf = ExpandEnv(hf1.File)
      break
     end
    end
    if par then break end
   end
   if par then
    par = hf..'::/'..par
   end
  end -- AAutoSearch
  if par == nil then
   -- Автоматическим поиском ключевое слово не найдено. Вопрос - где искать.
   if #HelpFiles == 1 then
    i=1
   elseif #HelpFiles > 1 then
    local list={}
    for j=1,#HelpFiles do
     list[j] = {text=HelpFiles[j].Alias or HelpFiles[j].File}
    end
    local Item,Pos=far.Menu({Title='Какую справку смотреть?'}, list)
    if Item then
     i = Pos
    end
   end
   if i > 0 then
    hf = ExpandEnv(HelpFiles[i].File) -- Имя help-файла
    if HelpFiles[i].Cmd then -- Явно указана команда для вызова
     cmd = HelpFiles[i].Cmd
     if HelpFiles[i].Par then -- Указаны параметры
      par = string.format(HelpFiles[i].Par, KeyString, hf)
     else -- Параметры не указаны; единственный параметр - имя файла справки
      par = hf
     end
    else -- Команда не указана. Для известных систем сам формирую.
     if _FileExt(hf) == '.hlp' then
      cmd = 'winhlp32'
      if KeyString ~= '' then
       par = string.format('-k %s %s', KeyString, hf)
      else
       par = hf
      end
     elseif _FileExt(hf) == '.chm' then
      cmd = 'hh'
      par = OpnChm(hf, KeyString)
      if par then
       par = hf..'::/'..par
      else
       par = hf
      end
     end
    end
   end -- if i > 0
  end -- if par == nil
  if not cmd then
   cmd = hf
  end;
  if cmd then
   win.ShellExecute(nil, nil, cmd, par)
  end
end -- GetHelp

Macro{
  id="1CA4FBB1-21EB-490E-B672-AF5004E7D722";
  area="Editor";
  key=S.ExecKey;
  description="Заменитель Active-Help - выполнение отмеченного через ShellExecute";
--  filemask="";
  flags="EVSelection";
--  priority=50;
--  sortpriority=50;
  condition=function()
    return true
  end;
  action=function()
   win.ShellExecute(nil, nil, Editor.SelValue)
  end;
}

Macro{
  id="E5B7615B-E825-44C4-A7CD-C90F3BEEAD92";
  area="Editor";
  key=S.HelpKey..' '..S.HelpKey..':Double'; -- CtrlF1, CtrlF1:Double
  description="Заменитель Active-Help - вызов справки по языку";
  condition=function()
    -- Составление списка настроек на помощь, подходящих к типу файла
    return HelpFList(editor.GetFileName())
  end;
  action=function()
   local function _GetWord()
    local Rez
    local SelSave = editor.GetSelection(nil)
    if SelSave then
     Rez = Editor.SelValue
    else
     Keys('SelWord')
     Rez = Editor.SelValue
     editor.Select(nil, F.BTYPE_NONE) -- гашение выделения
    end
    return Rez
   end -- _GetWord
   if akey(1,1):match(':Double') then
    GetHelp(_GetWord(), false) -- С блокировкой автопоиска
   else
    GetHelp(_GetWord(), true)
   end
  end;
}

Macro{
  id="C12A820A-2587-449A-A83A-49E85713D004";
  area="Dialog";
  --key=S.HelpKey..' '..S.HelpKey..':Double'; -- CtrlF1, CtrlF1:Double
  description="Заменитель Active-Help - справка по функции";
  condition=function()
    -- Составление списка настроек на помощь, подходящих к типу файла
    return HelpFList(editor.GetFileName())
  end;
  action=function(fn, Word, Auto)
   if HelpFList('*.'..fn) then
    AutoSearch = Auto
    GetHelp(Word, Auto) -- С блокировкой автопоиска
   end
  end;
}

Macro {
  id="F53B43B1-0D9C-47E8-B1D1-81C9086A9B82";
  area = "Common";
  key = S.SetupKey; -- ALtShiftF12
  description="Заменитель Active-Help - Настройка";
--  filemask="";
  flags = "EnableOutput";
--  priority=50;
--  sortpriority=50;
  condition=function()
    return true
  end;
  action = Active_Help_Setup;
}

CommandLine {
 description = "Заменитель Active-Help - вызов помощи из командной строки";
 --prefixes = "achlp:ac-hlp:achlpc:ac-hlp-c:";
 prefixes = S.Prefix..':'..S.Prefixc;
 action = function(prefix, text)
  local fn, keystr=text:match('(.-)%s(.*)')
  if not fn then
   fn=text
   keystr = nil
  end
  if mf.index(fn, '.') == -1 then -- Указано только расширение
   fn = '*.'..fn
  end;
  if HelpFList(fn) then
   if keystr then
    if mf.index(':'..S.Prefixc..':', ':'..prefix..':') >= 0 then
     GetHelp(keystr, false) -- НЕ автопоиск
    else
     GetHelp(keystr, true) -- Автопоиск
    end
   else
    GetHelp('', false)
   end
  end
 end;
}

MenuItem {
  description = "Active-Help Setup";
  menu   = "Config";
  guid   = "5FCE84E4-BA42-4F74-B9EF-2FBAB88FEA9A";
  text   = "BAX - Настройка Active-Help";
  action = function(OpenFrom,Item) Active_Help_Setup() end;      -- function
}

nfo.config      = Active_Help_Setup;